"""
health checks
"""
from __future__ import annotations

import asyncio
import logging
import time
from enum import Enum
from typing import Any, Callable, Type, Optional

from aspyx.di import injectable, Environment, inject_environment, on_init
from aspyx.reflection import Decorators, TypeDescriptor


def health_checks():
    """
    Instances of classes that are annotated with @injectable can be created by an Environment.
    """
    def decorator(cls):
        Decorators.add(cls, health_checks)

        #if not Providers.is_registered(cls):
        #    Providers.register(ClassInstanceProvider(cls, True, "singleton"))

        HealthCheckManager.types.append(cls)

        return cls

    return decorator

def health_check(name="", cache = 0, fail_if_slower_than = 0):
    """
    Methods annotated with `@on_init` will be called when the instance is created.
    """
    def decorator(func):
        Decorators.add(func, health_check, name, cache, fail_if_slower_than)
        return func

    return decorator

class HealthStatus(Enum):
    """
    A enum specifying the health status of a service. The values are:

    - `OK` service is healthy
    - `WARNING` service has some problems
    - `CRITICAL` service is unhealthy
    """
    OK = 1
    WARNING = 2
    ERROR = 3

    def __str__(self):
        return self.name


@injectable()
class HealthCheckManager:
    """
    The health manager is able to run all registered health checks and is able to return an overall status.
    """
    logger = logging.getLogger("aspyx.service.health")

    # local classes

    class Check:
        def __init__(self, name: str, cache: int, fail_if_slower_than: int, instance: Any, callable: Callable):
            self.name = name
            self.cache = cache
            self.callable = callable
            self.instance = instance
            self.fail_if_slower_than = fail_if_slower_than
            self.last_check = 0

            self.last_value : Optional[HealthCheckManager.Result] = None

        async def run(self, result: HealthCheckManager.Result):
            now = time.time()

            if self.cache > 0 and self.last_check is not None and now - self.last_check < self.cache:
                result.copy_from(self.last_value)
                return

            self.last_check = now
            self.last_value = result

            if asyncio.iscoroutinefunction(self.callable):
                await self.callable(self.instance, result)
            else:
                await asyncio.to_thread(self.callable, self.instance, result)

            spent = time.time() - now

            if result.status == HealthStatus.OK and 0 < self.fail_if_slower_than < spent:
                result.status = HealthStatus.ERROR
                result.details = f"spent {spent:.3f}s"


    class Result:
        def __init__(self, name: str):
            self.status = HealthStatus.OK
            self.name = name
            self.details = ""

        def copy_from(self, value: HealthCheckManager.Result):
            self.status  = value.status
            self.details = value.details

        def set_status(self, status: HealthStatus, details =""):
            self.status = status
            self.details = details

        def to_dict(self):
            result = {
                "name": self.name,
                "status": str(self.status),
            }

            if self.details:
                result["details"] = self.details

            return result

    class Health:
        def __init__(self, status: HealthStatus = HealthStatus.OK):
            self.status = status
            self.results : list[HealthCheckManager.Result] = []

        def to_dict(self):
            return {
                "status": str(self.status),
                "checks": [result.to_dict() for result in self.results]
            }

    # class data

    types : list[Type] = []

    # constructor

    def __init__(self):
        self.environment : Optional[Environment] = None
        self.checks: list[HealthCheckManager.Check] = []

    # check

    async def check(self) -> HealthCheckManager.Health:
        """
        run all registered health checks and return an overall result.
        Returns: the overall result.

        """
        self.logger.info("Checking health...")

        health = HealthCheckManager.Health()

        tasks = []
        for check in self.checks:
            result = HealthCheckManager.Result(check.name)
            health.results.append(result)
            tasks.append(check.run(result))

        await asyncio.gather(*tasks)

        for result in health.results:
            if result.status.value > health.status.value:
                health.status = result.status

        return health

    # public

    @inject_environment()
    def set_environment(self, environment: Environment):
        self.environment = environment

    @on_init()
    def setup(self):
        for type in self.types:
            descriptor = TypeDescriptor(type).for_type(type)
            instance = self.environment.get(type)

            for method in descriptor.get_methods():
                if method.has_decorator(health_check):
                    decorator = method.get_decorator(health_check)

                    name = decorator.args[0]
                    cache = decorator.args[1]
                    fail_if_slower_than = decorator.args[2]
                    if len(name) == 0:
                        name = method.get_name()

                    self.checks.append(HealthCheckManager.Check(name, cache, fail_if_slower_than, instance, method.method))



"""
jwt sample test
"""
import faulthandler
import time

from aspyx.di.configuration import inject_value
from aspyx.util import Logger

faulthandler.enable()

from typing import Optional

import logging
from abc import abstractmethod

from fastapi import Request as HttpRequest, HTTPException

from datetime import datetime, timezone

Logger.configure(default_level=logging.DEBUG, levels={
    "httpx": logging.ERROR,
    "aspyx.di": logging.ERROR,
    "aspyx.di.aop": logging.ERROR,
    "aspyx.service": logging.ERROR
})

from aspyx_service import Service, service, component, implementation, AbstractComponent, \
    Component, ChannelAddress, Server, health, RequestContext, HTTPXChannel, \
    AbstractAuthorizationFactory, AuthorizationManager, SessionManager, AuthorizationException, Session, \
    ServiceCommunicationException, TokenExpiredException, ServiceManager, TokenContext

from aspyx.reflection import Decorators, TypeDescriptor

from aspyx.di import injectable, order
from aspyx.di.aop import advice, around, Invocation, methods, classes


from .common import service_manager, TokenManager

# decorator

def secure():
    def decorator(cls):
        Decorators.add(cls, secure)

        return cls

    return decorator

# session

class UserSession(Session):
    # constructor

    def __init__(self, user: str, roles: list[str]):
        super().__init__()

        self.user = user
        self.roles = roles

# advice

def requires_role(role=""):
    """
    Methods decorated with `@requires_role` will only be allowed if the current user has the given role.
    """
    def decorator(func):
        Decorators.add(func, requires_role, role)

        return func

    return decorator


@injectable()
@order(1)
class RoleAuthorizationFactory(AbstractAuthorizationFactory):
    # local class

    class RoleAuthorization(AuthorizationManager.Authorization):
        # constructor

        def __init__(self, role: str):
            self.role = role

        # implement

        def check(self, invocation: Invocation):
            if not self.role in SessionManager.current(UserSession).roles:
                raise AuthorizationException(f"expected role {self.role}")
    # implement

    def compute_authorization(self, method_descriptor: TypeDescriptor.MethodDescriptor) -> Optional[AuthorizationManager.Authorization]:
        if method_descriptor.has_decorator(requires_role):
            role = method_descriptor.get_decorator(requires_role).args[0]
            return RoleAuthorizationFactory.RoleAuthorization(role)

        return None

@injectable()
@order(0)
class TokenAuthorizationFactory(AbstractAuthorizationFactory):
    # local class

    class TokenAuthorization(AuthorizationManager.Authorization):
        # constructor

        def __init__(self, session_manager: SessionManager, token_manager: TokenManager):
            self.session_manager = session_manager
            self.token_manager = token_manager

        # internal

        def extract_token_from_request(self, request: HttpRequest) -> str:
            auth_header = request.headers.get("Authorization")
            if not auth_header or not auth_header.startswith("Bearer "):
                raise HTTPException(status_code=401, detail="Missing token",
                                    headers={
                                        "WWW-Authenticate": 'Bearer error="invalid_token", error_description="missing token"'}
                                    )

            return auth_header.split(" ")[1]

        # implement

        def check(self, invocation: Invocation):
            http_request = RequestContext.get_request()

            if http_request is not None:
                token = self.extract_token_from_request(http_request)

                session = self.session_manager.get_session(token)
                if session is None:
                    # verify token

                    payload = self.token_manager.decode_jwt(token)

                    # create session

                    session = self.session_manager.create_session(payload)

                    expiry = datetime.fromtimestamp(payload["exp"], tz=timezone.utc)
                    self.session_manager.store_session(token, session, expiry)

                # set thread local

                SessionManager.set_session(session)

    # constructor

    def __init__(self, session_manager: SessionManager, token_manager: TokenManager):
        super().__init__()

        self.token_authorization = TokenAuthorizationFactory.TokenAuthorization(session_manager, token_manager)

    # implement

    def compute_authorization(self, method_descriptor: TypeDescriptor.MethodDescriptor) -> Optional[AuthorizationManager.Authorization]:
        return self.token_authorization

@advice
@injectable()
class RetryAdvice:
    # constructor

    def __init__(self, token_manager: TokenManager):
        self.max_attempts = 3
        self.backoff_base = 0.2
        self.token_manager = token_manager

    # inject

    @inject_value("channel.max_attempts", 3)
    def set_max_attempts(self, max_attempts: int):
        self.max_attempts = max_attempts

    @inject_value("channel.backoff_base", 0.2)
    def set_backoff_base(self, backoff_base: float):
        self.backoff_base = backoff_base

    # sending side

    def refresh_token_if_possible(self):
        refresh_token = TokenContext.get_refresh_token()
        if not refresh_token:
            raise TokenExpiredException("No refresh token available")

        new_access_token = self.token_manager.refresh_access_token(refresh_token)
        TokenContext.set(new_access_token)

    @around(methods().of_type(HTTPXChannel).named("request"))
    def retry_request(self, invocation: Invocation):
        for attempt in range(1, self.max_attempts + 1):
            try:
                return invocation.proceed()

            except AuthorizationException as e:
                raise

            except TokenExpiredException as e:
                if attempt == self.max_attempts:
                    raise
                self.refresh_token_if_possible()

            except ServiceCommunicationException:
                ServiceManager.logger.warning(f"Request failed ({invocation.func.__name__}), attempt {attempt}/{self.max_attempts}")

                if attempt == self.max_attempts:
                    raise

                time.sleep(self.backoff_base * (2 ** (attempt - 1)))

        # make the compiler happy

        return None

    @around(methods().of_type(HTTPXChannel).named("request_async"))
    async def retry_async_request(self, invocation: Invocation):
        for attempt in range(1, self.max_attempts + 1):
            try:
                return await invocation.proceed_async()

            except TokenExpiredException as e:
                if attempt == self.max_attempts:
                    raise

                self.refresh_token_if_possible()

            except ServiceCommunicationException:
                ServiceManager.logger.warning(f"Request failed ({invocation.func.__name__}), attempt {attempt}/{self.max_attempts}")

                if attempt == self.max_attempts:
                    raise

                time.sleep(self.backoff_base * (2 ** (attempt - 1)))

        return None

@advice
@injectable()
class AuthorizationAdvice:
    # constructor

    def __init__(self, authorization_manager: AuthorizationManager, session_manager: SessionManager):
        self.authorization_manager = authorization_manager

        session_manager.set_session_factory(lambda token: UserSession(user=token.get("sub"), roles=token.get("roles")))

    # internal

    def authorize(self, invocation: Invocation):
        try:
            self.authorization_manager.check(invocation)
        except AuthorizationException as e:
            ServiceManager.logger.warning(f"Authorization failed ({invocation.func.__name__}): {str(e)}")

            raise HTTPException(status_code=403, detail=str(e) + f" in function {invocation.func.__name__}")

    # aspects

    @around(methods().that_are_async().decorated_with(secure),
            methods().that_are_async().declared_by(classes().decorated_with(secure)))
    async def authorize_async(self, invocation: Invocation):
        try:
            self.authorize(invocation)

            return await invocation.proceed_async()
        finally:
            SessionManager.delete_session()

    @around(methods().that_are_sync().decorated_with(secure),
            methods().that_are_sync().declared_by(classes().decorated_with(secure)))
    def authorize_sync(self, invocation: Invocation):
        try:
            self.authorize(invocation)

            return invocation.proceed()
        finally:
            SessionManager.delete_session()

# some services

@service(description="login service")
class LoginService(Service):
    @abstractmethod
    def login(self, user: str, password: str) -> Optional[dict]:
        pass

    @abstractmethod
    def logout(self):
        pass

@service(description="secured service")
@secure()
class SecureService(Service):
    @abstractmethod
    def secured(self):
        pass

    @abstractmethod
    def secured_admin(self):
        pass

@component(name="login-component", services=[LoginService, SecureService])
class JWTComponent(Component): # pylint: disable=abstract-method
    pass

@implementation()
class LoginServiceImpl(LoginService):
    def __init__(self, token_manager: TokenManager):
        self.token_manager = token_manager
        self.users = {
            "hugo": {
                "username": "hugo",
                "password": "secret",
                "roles": ["user"]
            },
            "andi": {
                "username": "andi",
                "password": "secret",
                "roles": ["user", "admin"]
            }
        }

    def login(self, user: str, password: str) -> Optional[dict]:
        profile = self.users.get(user, None)
        if profile is not None and profile.get("password") == password:
            access_token = self.token_manager.create_access_token(user, profile.get("roles"))
            refresh_token = self.token_manager.create_refresh_token(user, profile.get("roles"))

            return {
                "access_token": access_token,
                "refresh_token": refresh_token,
            }

        return None

    def logout(self):
        pass

@implementation()
@secure()
class SecureServiceServiceImpl(SecureService):
    def __init__(self):
        pass

    def secured(self):
        session = SessionManager.current(UserSession)

    @requires_role("admin")
    def secured_admin(self):
        session = SessionManager.current(UserSession)


@implementation()
@health("/jwt-health")
class JWTComponentImpl(AbstractComponent, JWTComponent):
    # constructor

    def __init__(self):
        super().__init__()

    # implement

    def get_addresses(self, port: int) -> list[ChannelAddress]:
        return [
            ChannelAddress("dispatch-json", f"http://{Server.get_local_ip()}:{port}"),
            ChannelAddress("dispatch-msgpack", f"http://{Server.get_local_ip()}:{port}")
        ]

class TestLocalService():
    def test_login(self, service_manager):
        login_service = service_manager.get_service(LoginService, preferred_channel="dispatch-json")

        secure_service = service_manager.get_service(SecureService, preferred_channel="dispatch-json")

        try:
            secure_service.secured()
        except Exception as e:
            print(e)

        tokens = login_service.login("hugo", "secret")

        with TokenContext.use(tokens["access_token"], tokens["refresh_token"]):
            secure_service.secured()

            try:
                secure_service.secured_admin()
            except Exception as e:
                print(e)

            login_service.logout()

            TokenContext.clear()

            # now andi

            tokens = login_service.login("andi", "secret")

            TokenContext.set(tokens["access_token"], tokens["refresh_token"])

            secure_service.secured_admin()



"""
This module provides the core Aspyx service management framework allowing for service discovery and transparent remoting including multiple possible transport protocols.
"""

from aspyx.di import module

from .service import AuthorizationException, MissingTokenException, RemoteServiceException, ServiceCommunicationException, TokenException, TokenExpiredException, InvalidTokenException, component_services, ServiceException, Server, Channel, ComponentDescriptor, inject_service, ChannelAddress, ChannelInstances, ServiceManager, Component, Service, AbstractComponent, ComponentStatus, ComponentRegistry, implementation, health, component, service
from .channels import HTTPXChannel, DispatchJSONChannel, TokenContext
from .registries import ConsulComponentRegistry
from .server import FastAPIServer, RequestContext
from .healthcheck import health_checks, health_check, HealthCheckManager, HealthStatus
from .restchannel import RestChannel, post, get, put, delete, QueryParam, Body, rest
from .session import Session, SessionManager
from .authorization import AuthorizationManager, AbstractAuthorizationFactory, AuthorizationException


@module()
class ServiceModule:
    def __init__(self):
        pass

__all__ = [
    # service

    "ServiceManager",
    "ServiceModule",
    "ServiceException",
    "Server",
    "Component",
    "Service",
    "Channel",
    "AbstractComponent",
    "ComponentStatus",
    "ComponentDescriptor",
    "ComponentRegistry",
    "ChannelAddress",
    "ChannelInstances",
    "health",
    "component",
    "service",
    "implementation",
    "inject_service",
    "component_services",
    "RemoteServiceException",
    "ServiceCommunicationException",
    "TokenException",
    "TokenExpiredException",
    "InvalidTokenException",
    "MissingTokenException",
    "AuthorizationException",

    # authorization

    "AuthorizationManager",
    "AbstractAuthorizationFactory",
    "AuthorizationException",

    # session

    "Session",
    "SessionManager",

    # healthcheck

    "health_checks",
    "health_check",
    "HealthStatus",
    "HealthCheckManager",

    # serialization

   # "deserialize",

    # channel

    "HTTPXChannel",
    "DispatchJSONChannel",
    "TokenContext",

    # rest

    "RestChannel",
    "post",
    "get",
    "put",
    "delete",
    "rest",
    "QueryParam",
    "Body",

    # registries

    "ConsulComponentRegistry",
    "RequestContext",

    # server

    "FastAPIServer",
    "RequestContext"
]



"""
Common test stuff
"""
import logging
import time
from datetime import datetime, timedelta, timezone

import jwt
from fastapi import HTTPException
from jwt import ExpiredSignatureError, InvalidTokenError
from abc import abstractmethod
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, Optional, Any

import pytest
from pydantic import BaseModel

from aspyx.di.aop import advice, error, Invocation
from aspyx.exception import ExceptionManager, handle
#from aspyx.util import Logger
from aspyx_service import service, Service, component, Component, \
    implementation, health, AbstractComponent, ChannelAddress, inject_service, \
    FastAPIServer, Server, ServiceModule, ServiceManager, \
    HealthCheckManager, get, post, rest, put, delete, Body
from aspyx_service.service import LocalComponentRegistry, component_services, AuthorizationException
from aspyx.di import module, create, injectable, on_running
from aspyx.di.configuration import YamlConfigurationSource

# configure logging

#Logger.configure(default_level=logging.DEBUG, levels={
#    "httpx": logging.ERROR,
#    "aspyx.di": logging.INFO,
#    "aspyx.di.aop": logging.ERROR,
#    "aspyx.service": logging.ERROR
#})

# classes

class Pydantic(BaseModel):
    i : int
    f : float
    b: bool
    s: str

@dataclass
class Data:
    i: int
    f: float
    b: bool
    s: str

class PydanticAndData(BaseModel):
    p: Pydantic

@dataclass
class DataAndPydantic:
    d: Data

# jwt


SECRET_KEY = "your-secret-key"
ALGORITHM = "HS256"
ACCESS_TOKEN_EXPIRE_MINUTES = 30

class TokenManager:
    # constructor

    def __init__(self, secret: str, algorithm: str, access_token_expiry_minutes: int = 15, refresh_token_expiry_minutes: int = 60 * 24):
        self.secret = secret
        self.algorithm = algorithm
        self.access_token_expiry_minutes = access_token_expiry_minutes
        self.refresh_token_expiry_minutes = refresh_token_expiry_minutes

    # methods

    def create_jwt(self, subject: str, roles: list[str]) -> str:
        return self.create_access_token(subject, roles)

    def create_access_token(self, subject: str, roles: list[str]) -> str:
        now = datetime.now(tz=timezone.utc)
        expiry = now + timedelta(minutes=self.access_token_expiry_minutes)

        payload = {
            "sub": subject,
            "roles": roles,
            "exp": int(expiry.timestamp()),
            "iat": int(now.timestamp()),
            "type": "access"
        }

        return jwt.encode(payload, self.secret, algorithm=self.algorithm)

    def create_refresh_token(self, subject: str, roles: list[str]) -> str:
        now = datetime.now(tz=timezone.utc)
        expiry = now + timedelta(minutes=self.refresh_token_expiry_minutes)

        payload = {
            "sub": subject,
            "roles": roles,
            "exp": int(expiry.timestamp()),
            "iat": int(now.timestamp()),
            "type": "refresh"
        }

        return jwt.encode(payload, self.secret, algorithm=self.algorithm)

    def refresh_access_token(self, refresh_token: str) -> str:
        payload = self.decode_jwt(refresh_token)

        if payload.get("type") != "refresh":
            raise AuthorizationException("Expected a refresh token")

        subject = payload.get("sub")
        if not subject:
            raise AuthorizationException("Missing subject in refresh token")

        roles = payload.get("roles")

        return self.create_access_token(subject, roles)

    def decode_jwt(self, token: str) -> dict[str, Any]:
        try:
            return jwt.decode(token, self.secret, algorithms=[self.algorithm])
        except ExpiredSignatureError:
            raise HTTPException(status_code=401,
                                detail="Token has expired",
                                headers={"WWW-Authenticate": 'Bearer error="invalid_token", error_description="The token has expired"'}
                                )
        except InvalidTokenError:
            raise HTTPException(
                status_code=401,
                detail="Invalid token",
                headers={"WWW-Authenticate": 'Bearer error="invalid_token", error_description="The token is invalid"'}
            )


# service

@service(name="test-service", description="cool")
class TestService(Service):
    @abstractmethod
    def hello(self, message: str) -> str:
        pass

    @abstractmethod
    def throw(self, message: str) -> str:
        pass

    @abstractmethod
    def data(self, data: Data) -> Data:
        pass

    @abstractmethod
    def pydantic(self, data: Pydantic) -> Pydantic:
        pass

@service(name="test-async-service", description="cool")
class TestAsyncService(Service):
    @abstractmethod
    async def hello(self, message: str) -> str:
        pass

    @abstractmethod
    async def data(self, data: Data) -> Data:
        pass

    @abstractmethod
    async def pydantic(self, data: Pydantic) -> Pydantic:
        pass

@service(name="test-rest-service", description="cool")
@rest("/api")
class TestRestService(Service):
    @abstractmethod
    @get("/hello/{message}")
    def get(self, message: str) -> str:
        pass

    @put("/hello/{message}")
    def put(self, message: str) -> str:
        pass

    @post("/hello/{message}")
    def post_pydantic(self, message: str, data: Body(Pydantic)) -> Pydantic:
        pass

    @post("/hello/{message}")
    def post_data(self, message: str, data: Body(Data)) -> Data:
        pass

    @delete("/hello/{message}")
    def delete(self, message: str) -> str:
        pass

@service(name="test-async-rest-service", description="cool")
@rest("/async-api")
class TestAsyncRestService(Service):
    @abstractmethod
    @get("/hello/{message}")
    async def get(self, message: str) -> str:
        pass

    @put("/hello/{message}")
    async def put(self, message: str) -> str:
        pass

    @post("/hello/{message}")
    async def post_pydantic(self, message: str, data: Body(Pydantic)) -> Pydantic:
        pass

    @post("/hello/{message}")
    async def post_data(self, message: str, data: Body(Data)) -> Data:
        pass

    @delete("/hello/{message}")
    async def delete(self, message: str) -> str:
        pass

@component(services =[
    TestService,
    TestAsyncService,
    TestRestService,
    TestAsyncRestService
])
class TestComponent(Component): # pylint: disable=abstract-method
    pass

# implementation classes

@implementation()
class TestServiceImpl(TestService):
    def __init__(self):
        pass

    def hello(self, message: str) -> str:
        return message

    def throw(self, message: str) -> str:
        raise Exception(message)

    def data(self, data: Data) -> Data:
        return data

    def pydantic(self, data: Pydantic) -> Pydantic:
        return data

@implementation()
class TestAsyncServiceImpl(TestAsyncService):
    def __init__(self):
        pass

    async def hello(self, message: str) -> str:
        return message

    async def data(self, data: Data) -> Data:
        return data

    async def pydantic(self, data: Pydantic) -> Pydantic:
        return data

@implementation()
class TestRestServiceImpl(TestRestService):
    def __init__(self):
        pass

    def get(self, message: str) -> str:
        return message

    def put(self, message: str) -> str:
        return message

    def post_pydantic(self, message: str, data: Pydantic) -> Pydantic:
        return data

    def post_data(self, message: str, data: Data) -> Data:
        return data

    def delete(self, message: str) -> str:
        return message

@implementation()
class TestAsyncRestServiceImpl(TestAsyncRestService):
    def __init__(self):
        pass

    async def get(self, message: str) -> str:
        return message

    async def put(self, message: str) -> str:
        return message

    async def post_pydantic(self, message: str, data: Pydantic) -> Pydantic:
        return data

    async def post_data(self, message: str, data: Data) -> Data:
        return data

    async def delete(self, message: str) -> str:
        return message

@implementation()
@health("/health")
@advice
class TestComponentImpl(AbstractComponent, TestComponent):
    # constructor

    def __init__(self):
        super().__init__()

        self.health_check_manager : Optional[HealthCheckManager] = None
        self.exception_manager = ExceptionManager()

    # exception handler

    @handle()
    def handle_exception(self, exception: Exception):
        print("caught exception!")
        return exception

    # aspects

    @error(component_services(TestComponent))
    def catch(self, invocation: Invocation):
        return self.exception_manager.handle(invocation.exception)

    # lifecycle

    @on_running()
    def setup_exception_handlers(self):
        self.exception_manager.collect_handlers(self)

    # implement

    async def get_health(self) -> HealthCheckManager.Health:
        return HealthCheckManager.Health()

    def get_addresses(self, port: int) -> list[ChannelAddress]:
        return [
            ChannelAddress("rest", f"http://{Server.get_local_ip()}:{port}"),
            ChannelAddress("dispatch-json", f"http://{Server.get_local_ip()}:{port}"),
            ChannelAddress("dispatch-msgpack", f"http://{Server.get_local_ip()}:{port}")
        ]

    def startup(self) -> None:
        print("### startup")

    def shutdown(self) -> None:
        print("### shutdown")

@injectable(eager=False)
class Foo:
    def __init__(self):
        self.service = None

    @inject_service(preferred_channel="local")
    def set_service(self, service: TestService):
        self.service = service

# module

@module(imports=[ServiceModule])
class Module:
    def __init__(self):
        pass

    @create()
    def create_token_manager(self) -> TokenManager:
        return TokenManager(SECRET_KEY, ALGORITHM, access_token_expiry_minutes = 15, refresh_token_expiry_minutes = 60 * 24)

    @create()
    def create_yaml_source(self) -> YamlConfigurationSource:
        return YamlConfigurationSource(f"{Path(__file__).parent}/config.yaml")

    @create()
    def create_registry(self, source: YamlConfigurationSource) -> LocalComponentRegistry:
        return LocalComponentRegistry()

# main

def start_server() -> ServiceManager:
    server = FastAPIServer.boot(module=Module, host="0.0.0.0", port=8000)

    service_manager = server.service_manager
    descriptor = service_manager.get_descriptor(TestComponent).get_component_descriptor()

    # Give the server a second to start

    print("wait for server to start")
    while True:
        addresses = service_manager.component_registry.get_addresses(descriptor)
        if addresses:
            break

        print("zzz...")
        time.sleep(1)

    print("server running")

    return service_manager

manager : Optional[ServiceManager] = None

@pytest.fixture(scope="session")
def service_manager():
    global manager

    if  manager is None:
        manager = start_server()  # start server

    yield manager



"""
FastAPI server implementation for the aspyx service framework.
"""
import atexit
import inspect
import threading
from typing import Type, Optional, Callable, Any

from fastapi.responses import JSONResponse
import msgpack
import uvicorn

from fastapi import FastAPI, APIRouter, Request as HttpRequest, Response as HttpResponse, HTTPException
import contextvars

from starlette.middleware.base import BaseHTTPMiddleware

from aspyx.di import Environment, injectable, on_init, inject_environment
from aspyx.reflection import TypeDescriptor, Decorators

from .service import ComponentRegistry
from .healthcheck import HealthCheckManager

from .serialization import get_deserializer

from .service import Server, ServiceManager
from .channels import Request, Response, TokenContext

from .restchannel import get, post, put, delete, rest

class RequestContext:
    """
    A request context is used to remember the current http request in the current thread
    """
    request_var = contextvars.ContextVar("request")

    @classmethod
    def get_request(cls) -> Request:
        """
        Return the current http request

        Returns:
            the current http request
        """
        return cls.request_var.get()

    # constructor

    def __init__(self, app):
        self.app = app

    async def __call__(self, scope, receive, send):
        if scope["type"] != "http":
            await self.app(scope, receive, send)
            return

        request = HttpRequest(scope)
        token = self.request_var.set(request)
        try:
            await self.app(scope, receive, send)
        finally:
            self.request_var.reset(token)

class TokenMiddleware(BaseHTTPMiddleware):
    async def dispatch(self, request: Request, call_next):
        access_token = request.cookies.get("access_token") or request.headers.get("Authorization")
        #refresh_token = request.cookies.get("refresh_token")

        if access_token:
            TokenContext.set(access_token)#, refresh_token)

        try:
            return await call_next(request)
        finally:
            TokenContext.clear()

def create_server() -> FastAPI:
    server = FastAPI()

    server.add_middleware(RequestContext)
    server.add_middleware(TokenMiddleware)

    return server

@injectable()
class FastAPIServer(Server):
    """
    A server utilizing fastapi framework.
    """

    fast_api = create_server()

    # class methods

    @classmethod
    def boot(cls, module: Type, host="0.0.0.0", port=8000, start = True) -> 'FastAPIServer':
        """
        boot the DI infrastructure of the supplied module and optionally start a fastapi thread given the url
        Args:
            module: the module to initialize the environment
            host: listen address
            port: the port

        Returns:
            thr created server
        """
        cls.port = port

        environment = Environment(module)

        server = environment.get(FastAPIServer)

        if start:
            server.start_fastapi(host)

        return server

    # constructor

    def __init__(self, service_manager: ServiceManager, component_registry: ComponentRegistry):
        super().__init__()

        self.environment : Optional[Environment] = None
        self.service_manager = service_manager
        self.component_registry = component_registry

        self.host = "localhost"
        self.server_thread = None

        self.router = APIRouter()

        # cache

        self.deserializers: dict[str, list[Callable]] = {}

        # that's the overall dispatcher

        self.router.post("/invoke")(self.invoke)

    # inject

    @inject_environment()
    def set_environment(self, environment: Environment):
        self.environment = environment

    @on_init()
    def on_init(self):
        self.service_manager.startup(self)

        # add routes

        self.add_routes()
        self.fast_api.include_router(self.router)

        #for route in self.fast_api.routes:
        #    print(f"{route.name}: {route.path} [{route.methods}]")

        # add cleanup hook

        def cleanup():
            self.service_manager.shutdown()

        atexit.register(cleanup)

    # private

    def add_routes(self):
        """
        add everything that looks like an http endpoint
        """
        for descriptor in self.service_manager.descriptors.values():
            if not descriptor.is_component() and descriptor.is_local():
                prefix = ""

                type_descriptor = TypeDescriptor.for_type(descriptor.type)
                instance = self.environment.get(descriptor.implementation)

                if type_descriptor.has_decorator(rest):
                    prefix = type_descriptor.get_decorator(rest).args[0]

                for method in type_descriptor.get_methods():
                    decorator = next((decorator for decorator in Decorators.get(method.method) if decorator.decorator in [get, put, post, delete]), None)
                    if decorator is not None:
                        self.router.add_api_route(
                            path=prefix + decorator.args[0],
                            endpoint=getattr(instance, method.get_name()),
                            methods=[decorator.decorator.__name__],
                            name=f"{descriptor.get_component_descriptor().name}.{descriptor.name}.{method.get_name()}",
                            response_model=method.return_type,
                        )

    def start_fastapi(self, host: str):
        """
        start the fastapi server in a thread
        """
        self.host = host

        config = uvicorn.Config(self.fast_api, host=self.host, port=self.port, access_log=False)
        server = uvicorn.Server(config)

        thread = threading.Thread(target=server.run, daemon=True)
        thread.start()

        return thread

    def get_deserializers(self, service: Type, method):
        deserializers = self.deserializers.get(method, None)
        if deserializers is None:
            descriptor = TypeDescriptor.for_type(service).get_method(method.__name__)

            deserializers = [get_deserializer(type) for type in descriptor.param_types]
            self.deserializers[method] = deserializers

        return deserializers

    def deserialize_args(self, args: list[Any], type: Type, method: Callable) -> list:
        #args = list(request.args)

        deserializers = self.get_deserializers(type, method)

        for i, arg in enumerate(args):
            args[i] = deserializers[i](arg)

        return args

    async def invoke(self, http_request: HttpRequest):
        content_type = http_request.headers.get("content-type", "")

        content = "json"
        if "application/msgpack" in content_type:
            content = "msgpack"
            raw_data = await http_request.body()
            data = msgpack.unpackb(raw_data, raw=False)
        elif "application/json" in content_type:
            data = await http_request.json()
        else:
            return HttpResponse(
                content="Unsupported Content-Type",
                status_code=415,
                media_type="text/plain"
            )

        request = data

        if content == "json":
            return await self.dispatch(http_request, request)
        else:
            return HttpResponse(
                content=msgpack.packb(await self.dispatch(http_request, request), use_bin_type=True),
                media_type="application/msgpack"
            )

    async def dispatch(self, http_request: HttpRequest, request: dict) :
        ServiceManager.logger.debug("dispatch request %s", request["method"])

        # <comp>:<service>:<method>

        parts = request["method"].split(":")

        #component = parts[0]
        service_name = parts[1]
        method_name = parts[2]

        service_descriptor = ServiceManager.descriptors_by_name[service_name]
        service = self.service_manager.get_service(service_descriptor.type, preferred_channel="local")

        method = getattr(service, method_name)

        args = self.deserialize_args(request["args"], service_descriptor.type, method)
        try:
            if inspect.iscoroutinefunction(method):
                result = await method(*args)
            else:
                result = method(*args)

            return Response(result=result, exception=None).model_dump()

        except HTTPException as e:
            raise

        except Exception as e:
            return Response(result=None, exception=str(e)).model_dump()

    # override

    def route(self, url: str, callable: Callable):
        self.router.get(url)(callable)

    def route_health(self, url: str, callable: Callable):
        async def get_health_response():
            health : HealthCheckManager.Health = await callable()

            return JSONResponse(
                status_code= self.component_registry.map_health(health),
                content = health.to_dict()
            )

        self.router.get(url)(get_health_response)



"""
session related module
"""
import contextvars
from typing import Type, Optional, Callable, Any, TypeVar
from datetime import datetime, timezone
from cachetools import TTLCache

from aspyx.di import injectable
from aspyx.threading import ThreadLocal


class Session:
    """
    Base class for objects covers data related to a server side session.
    """
    def __init__(self):
        pass

T = TypeVar("T")

@injectable()
class SessionManager:
    """
    A SessionManager controls the lifecycle of sessions and is responsible to establish a session thread local.
    """
    #current_session = ThreadLocal[Session]()
    current_session  = contextvars.ContextVar("session")

    @classmethod
    def current(cls, type: Type[T]) -> T:
        """
        return the current session associated with the thread
        Args:
            type:  the session type

        Returns:
            the current session
        """
        return cls.current_session.get()

    @classmethod
    def set_session(cls, session: Session) -> None:
        """
        set the current session in the thread context
        Args:
            session: the session
        """
        cls.current_session.set(session)

    @classmethod
    def delete_session(cls) -> None:
        """
        delete the current session
        """
        cls.current_session.set(None)#clear()

    # constructor

    def __init__(self):
        self.sessions = TTLCache(maxsize=1000, ttl=3600)
        self.session_creator : Optional[Callable[[Any], Session]] = None

    # public

    def set_session_factory(self, callable: Callable[..., Session]) -> None:
        """
        set a factory function that will be used to create a concrete session
        Args:
            callable: the function
        """
        self.session_creator = callable

    def create_session(self, *args, **kwargs) -> Session:
        """
        create a session given the argument s(usually a token, etc.)
        Args:
            args: rest args
            kwargs: keyword args

        Returns:
            the new session
        """
        return self.session_creator(*args, **kwargs)

    def store_session(self, token: str, session: Session, expiry: datetime):
        now = datetime.now(timezone.utc)
        ttl_seconds = max(int((expiry - now).total_seconds()), 0)
        self.sessions[token] = (session, ttl_seconds)

    def get_session(self, token: str) -> Optional[Session]:
        value = self.sessions.get(token)
        if value is None:
            return None

        session, ttl = value
        return session



"""
Tests
"""

from .common import TestService, TestRestService, Pydantic, Data, service_manager, Foo

pydantic = Pydantic(i=1, f=1.0, b=True, s="s")
data = Data(i=1, f=1.0, b=True, s="s")


class TestLocalService():
    def test_local(self, service_manager):
        test_service = service_manager.get_service(TestService, preferred_channel="local")

        result = test_service.hello("hello")
        assert result == "hello"

        result_data = test_service.data(data)
        assert result_data == data

        result_pydantic = test_service.pydantic(pydantic)
        assert result_pydantic == pydantic

    def test_throw(self, service_manager):
        test_service = service_manager.get_service(TestService, preferred_channel="local")

        try:
            test_service.throw("hello")
        except Exception as e:
            print(e)

    def test_inject(self, service_manager):
        test = service_manager.environment.get(Foo)

        assert test.service is not None

class TestSyncRemoteService:
    def test_dispatch_json(self, service_manager):
        test_service = service_manager.get_service(TestService, preferred_channel="dispatch-json")

        result = test_service.hello("hello")
        assert result == "hello"

        result_data = test_service.data(data)
        assert result_data == data

        result_pydantic = test_service.pydantic(pydantic)
        assert result_pydantic == pydantic

    def xtest_dispatch_msgpack(self, service_manager):
        test_service = service_manager.get_service(TestService, preferred_channel="dispatch-msgpack")

        result = test_service.hello("hello")
        assert result == "hello"

        result_data = test_service.data(data)
        assert result_data == data

        result_pydantic = test_service.pydantic(pydantic)
        assert result_pydantic == pydantic

    def test_dispatch_rest(self, service_manager):
        test_service = service_manager.get_service(TestRestService, preferred_channel="rest")

        result = test_service.get("hello")
        assert result == "hello"

        result = test_service.put("hello")
        assert result == "hello"

        result = test_service.delete("hello")
        assert result == "hello"

        # data and pydantic

        result_pydantic = test_service.post_pydantic("message", pydantic)
        assert result_pydantic == pydantic

        result_data= test_service.post_data("message", data)
        assert result_data == data



import logging
import os

from aspyx.util import Logger
from aspyx_service import FastAPIServer
from server import  ServerModule

Logger.configure(default_level=logging.DEBUG, levels={
    "httpx": logging.ERROR,
    "aspyx.di": logging.ERROR,
    "aspyx.di.aop": logging.ERROR,
    "aspyx.service": logging.ERROR
})

PORT = int(os.getenv("FAST_API_PORT", 8000))

FastAPIServer.boot(module=ServerModule, host="0.0.0.0", port=PORT, start = False)

app = FastAPIServer.fast_api

if __name__ == "__main__":
    import uvicorn
    uvicorn.run("main:app", host="0.0.0.0", port=PORT, reload=True, log_level="warning", access_log=False)


"""
Tests
"""
import asyncio
import threading
import time
import logging

from typing import Callable, TypeVar, Type, Awaitable, Any, Dict, cast

from consul import Consul

from aspyx.di import module, Environment, create
from aspyx.di.aop import advice, around, methods, Invocation
from aspyx.util import Logger

from aspyx_service import ConsulComponentRegistry
from aspyx_service.service import ServiceManager, ComponentRegistry, Channel

Logger.configure(default_level=logging.DEBUG, levels={
    "httpx": logging.ERROR,
    "aspyx.di": logging.ERROR,
    "aspyx.di.aop": logging.ERROR,
    "aspyx.service": logging.ERROR
})

from client import TestService, TestRestService, Pydantic, Data, TestAsyncRestService, TestAsyncService, ClientModule

# main

@advice
class ChannelAdvice:
    def __init__(self):
        pass

    @around(methods().named("customize").of_type(Channel))
    def customize_channel(self, invocation: Invocation):
        channel = cast(Channel, invocation.args[0])

        channel.select_round_robin() # or select_first_url()

        return invocation.proceed()

@module(imports=[ClientModule])
class TestModule:
    def __init__(self):
        pass

    @create()
    def create_registry(self) -> ComponentRegistry:
        return ConsulComponentRegistry(port=8000, consul=Consul(host="localhost", port=8500))

def boot() -> ServiceManager:
    environment = Environment(TestModule)

    service_manager = environment.get(ServiceManager)

    return service_manager

T = TypeVar("T")

# main

lorem_ipsum = "Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua"

pydantic = Pydantic(i=1, f=1.0, b=True, s="s",
                    str0=lorem_ipsum,
                    str1=lorem_ipsum,
                    str2=lorem_ipsum,
                    str3=lorem_ipsum,
                    str4=lorem_ipsum,
                    str5=lorem_ipsum,
                    str6=lorem_ipsum,
                    str7=lorem_ipsum,
                    str8=lorem_ipsum,
                    str9=lorem_ipsum
                    )
data = Data(i=1, f=1.0, b=True, s="s",
                    str0=lorem_ipsum,
                    str1=lorem_ipsum,
                    str2=lorem_ipsum,
                    str3=lorem_ipsum,
                    str4=lorem_ipsum,
                    str5=lorem_ipsum,
                    str6=lorem_ipsum,
                    str7=lorem_ipsum,
                    str8=lorem_ipsum,
                    str9=lorem_ipsum
            )

def run_loops(name: str, loops: int, type: Type[T], instance: T,  callable: Callable[[T], None]):
    start = time.perf_counter()
    for _ in range(loops):
        callable(instance)

    end = time.perf_counter()
    avg_ms = ((end - start) / loops) * 1000

    print(f"run {name}, loops={loops}: avg={avg_ms:.3f} ms")

async def run_async_loops(name: str, loops: int, type: Type[T], instance: T,  callable: Callable[[T], Awaitable[Any]]):
    start = time.perf_counter()
    for _ in range(loops):
        await callable(instance)

    end = time.perf_counter()
    avg_ms = ((end - start) / loops) * 1000

    print(f"run {name}, loops={loops}: avg={avg_ms:.3f} ms")

def run_threaded_async_loops(name: str, loops: int, n_threads: int,  type: Type[T], instance: T,  callable: Callable[[T], Awaitable[Any]]):
    threads = []

    def worker(thread_id: int):
        #print(f"worker {thread_id} running on thread {threading.current_thread().name}")

        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        async def run():
            for i in range(loops):
                await callable(instance)

        loop.run_until_complete(run())
        loop.close()

    start = time.perf_counter()

    for t_id in range(0, n_threads):
        thread = threading.Thread(target=worker, args=(t_id,))
        threads.append(thread)
        thread.start()

    for thread in threads:
        thread.join()

    end = time.perf_counter()
    took = (end - start) * 1000
    avg_ms = ((end - start) / (n_threads * loops)) * 1000

    print(f"{name} {loops} in {n_threads} threads: {took} ms, avg: {avg_ms}ms")

def run_threaded_sync_loops(name: str, loops: int, n_threads: int,  type: Type[T], instance: T,  callable: Callable[[T], Any]):
    threads = []

    def worker(thread_id: int):
        #print(f"worker {thread_id} running on thread {threading.current_thread().name}")

        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        async def run():
            for i in range(loops):
                callable(instance)

        loop.run_until_complete(run())
        loop.close()

    start = time.perf_counter()

    for t_id in range(0, n_threads):
        thread = threading.Thread(target=worker, args=(t_id,))
        threads.append(thread)
        thread.start()

    for thread in threads:
        thread.join()

    end = time.perf_counter()
    took = (end - start) * 1000
    avg_ms = ((end - start) / (n_threads * loops)) * 1000

    print(f"{name} {loops} in {n_threads} threads: {took} ms, avg: {avg_ms}ms")

manager = boot()

async def main():
    print("start tests...")

    # get service manager

    loops = 1000

    # tests

    run_loops("rest", loops, TestRestService, manager.get_service(TestRestService, preferred_channel="rest"), lambda service: service.get("world"))
    run_loops("json", loops, TestService, manager.get_service(TestService, preferred_channel="dispatch-json"), lambda service: service.hello("world"))
    run_loops("msgpack", loops, TestService, manager.get_service(TestService, preferred_channel="dispatch-msgpack"), lambda service: service.hello("world"))

    # pydantic

    run_loops("rest & pydantic", loops, TestRestService, manager.get_service(TestRestService, preferred_channel="rest"), lambda service: service.post_pydantic("hello", pydantic))
    run_loops("json & pydantic", loops, TestService, manager.get_service(TestService, preferred_channel="dispatch-json"), lambda service: service.pydantic(pydantic))
    run_loops("msgpack & pydantic", loops, TestService, manager.get_service(TestService, preferred_channel="dispatch-msgpack"), lambda service: service.pydantic(pydantic))

    # data class

    run_loops("rest & data", loops, TestRestService, manager.get_service(TestRestService, preferred_channel="rest"),
              lambda service: service.post_data("hello", data))
    run_loops("json & data", loops, TestService,
              manager.get_service(TestService, preferred_channel="dispatch-json"),
              lambda service: service.data(data))
    run_loops("msgpack & data", loops, TestService,
              manager.get_service(TestService, preferred_channel="dispatch-msgpack"),
              lambda service: service.data(data))

    # async

    await run_async_loops("async rest", loops, TestAsyncRestService, manager.get_service(TestAsyncRestService, preferred_channel="rest"),
              lambda service: service.get("world"))
    await run_async_loops("async json", loops, TestAsyncService, manager.get_service(TestAsyncService, preferred_channel="dispatch-json"),
              lambda service: service.hello("world"))
    await run_async_loops("async msgpack", loops, TestAsyncService, manager.get_service(TestAsyncService, preferred_channel="dispatch-msgpack"),
              lambda service: service.hello("world"))

    # pydantic

    await run_async_loops("async rest & pydantic", loops, TestAsyncRestService, manager.get_service(TestAsyncRestService, preferred_channel="rest"),
              lambda service: service.post_pydantic("hello", pydantic))
    await run_async_loops("async json & pydantic", loops, TestAsyncService,
              manager.get_service(TestAsyncService, preferred_channel="dispatch-json"),
              lambda service: service.pydantic(pydantic))
    await run_async_loops("async msgpack & pydantic", loops, TestAsyncService,
              manager.get_service(TestAsyncService, preferred_channel="dispatch-msgpack"),
              lambda service: service.pydantic(pydantic))

    # data class

    # pydantic

    await run_async_loops("async rest & data", loops, TestAsyncRestService, manager.get_service(TestAsyncRestService, preferred_channel="rest"),
              lambda service: service.post_data("hello", data))
    await run_async_loops("async json & data", loops, TestAsyncService,
              manager.get_service(TestAsyncService, preferred_channel="dispatch-json"),
              lambda service: service.data(data))
    await run_async_loops("async msgpack & data", loops, TestAsyncService,
              manager.get_service(TestAsyncService, preferred_channel="dispatch-msgpack"),
              lambda service: service.data(data))

    # a real thread test

    # sync

    run_threaded_sync_loops("threaded sync json, 1 thread", loops, 1, TestService,
                             manager.get_service(TestService, preferred_channel="dispatch-json"),
                             lambda service: service.hello("world"))
    run_threaded_sync_loops("threaded sync json, 2 thread", loops, 2, TestService,
                             manager.get_service(TestService, preferred_channel="dispatch-json"),
                             lambda service: service.hello("world"))
    run_threaded_sync_loops("threaded sync json, 4 thread", loops, 4, TestService,
                             manager.get_service(TestService, preferred_channel="dispatch-json"),
                             lambda service: service.hello("world"))
    run_threaded_sync_loops("threaded sync json, 8 thread", loops, 8, TestService,
                             manager.get_service(TestService, preferred_channel="dispatch-json"),
                             lambda service: service.hello("world"))
    run_threaded_sync_loops("threaded sync json, 16 thread", loops, 16, TestService,
                             manager.get_service(TestService, preferred_channel="dispatch-json"),
                             lambda service: service.hello("world"))

    # async

    run_threaded_async_loops("threaded async json, 1 thread", loops, 1, TestAsyncService,
                             manager.get_service(TestAsyncService, preferred_channel="dispatch-json"),
                             lambda service: service.hello("world") )
    run_threaded_async_loops("threaded async json, 2 thread", loops, 2, TestAsyncService,
                             manager.get_service(TestAsyncService, preferred_channel="dispatch-json"),
                             lambda service: service.hello("world"))
    run_threaded_async_loops("threaded async json, 4 thread", loops, 4, TestAsyncService,
                             manager.get_service(TestAsyncService, preferred_channel="dispatch-json"),
                             lambda service: service.hello("world"))
    run_threaded_async_loops("threaded async json, 8 thread", loops, 8, TestAsyncService,
                             manager.get_service(TestAsyncService, preferred_channel="dispatch-json"),
                             lambda service: service.hello("world"))
    run_threaded_async_loops("threaded async json, 16 thread", loops, 16, TestAsyncService,
                             manager.get_service(TestAsyncService, preferred_channel="dispatch-json"),
                             lambda service: service.hello("world"))

if __name__ == "__main__":
    asyncio.run(main())